%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% File    : leex.erl
%% Author  : Robert Virding
%% Purpose : Lexical analyser generator for Erlang.

-module(leex).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').

-compile(export_all).

-import(lists, [append/2,reverse/1,keymember/3,keysort/2]).

-import(ordsets, [add_element/2,is_element/2,subtract/2,union/2]).

-include("erl_compile.hrl").

%% To construct the NFA we use Thompson's construction with the slight
%% modification that a state can have any number of epsilon transitions
%% from it. This allows us to generate us to generate one "or" state
%% instead of a chain of states. We still, however, only allow one
%% character transition from a state.
%%
%% There is only one real optimistation we use when generating the NFA.
%% For character sets there is a special construction to denote a range.
%% This is used instead of seperate states for each character and saves
%% us from generating uninteresting extra states in both the NFA and DFA.
%% We are careful to calculate true ranges, not just what is in the
%% character set. The DFA generation knows about this and is also careful
%% to "pull it apart" correctly.
%%
%% The NFA is a list of states which have the format:
%%
%%      {nfa,Number,Transitions,Action}
%%
%% where Transitions is a list of
%%
%%      {Type,State} where Type = epsilon | <char> | {range,Cf,Cl}
%%
%% The DFA is a list of states which have the format:
%%
%%      {dfa,Number,NFAStates,Closure,Transitions,Action}
%%
%% where Transitions is a list of
%%
%%      {Type,State} where Type = <char> | {range,Cf,Cl}
%%
%% The NFAs for the regular expressions are generated in the same order
%% as the they have been given. This implies that the numbers of the
%% accepting state for each RE are in the same order as the REs. When
%% extracting the action for a DFA state from its NFA states.

gen(In, Out) ->
    InFile = lists:concat([In,".xrl"]),
    OutFile = lists:concat([Out,".erl"]),
    case parse_file(InFile) of
	{ok,REAs,Actions,Code} ->
	    {NFA,NF} = build_combined_nfa(REAs),
	    io:fwrite("NFA contains ~w states, ", [length(NFA)]),
	    {DFA0,DF0} = build_dfa(NFA, NF),
	    io:fwrite("DFA contains ~w states, ", [length(DFA0)]),
	    {DFA,DF} = minimise_dfa(DFA0, DF0),
	    io:fwrite("minimised to ~w states.~n", [length(DFA)]),
	    out_file(OutFile, Out, DFA, DF, Actions, Code),
	    ok;
	{error,Error} ->
	    io:put_chars([$\n,gcc_error(InFile, Error),$\n]),
	    error
    end.

format_error({open,F}) ->
    ["error opening ",io_lib:write_string(F)];
format_error(bad_macro) -> "bad macro";
format_error(missing_rules) -> "missing rules";
format_error(bad_rule) -> "bad rule";
format_error({regexp,S}) ->
    ["bad regexp ",io_lib:write_string(S)];
format_error({after_regexp,S}) ->
    ["bad code after regexp ",io_lib:write_string(S)].

gcc_error(File, {Line,Mod,Error}) ->
    io_lib:format("~s:~w: ~s", [File,Line,apply(Mod, format_error, [Error])]);
gcc_error(File, {Mod,Error}) ->
    io_lib:format("~s: ~s", [File,apply(Mod, format_error, [Error])]).

%% parse_file(InFile)
%%  Read and parse the file InFile. Return:
%%
%%      {REAs,Actions,Code} or 'error' on any error
%%  where
%%      REAs is a list {RegExp,ActionNo}
%%      Actions is a list of {ActionNo,ActionString}
%%      Code is a list of bytes
%%
%%  After each section of the file has been parsed we directly call the
%%  next section. This is done when we detect a line we don't recognise
%%  in the current section. The file format is very simple and Erlang
%%  token based, we allow empty lines and Erlang style comments.

parse_file(InFile) ->
    case file:open(InFile, read) of
	{ok,Ifile} ->
	    io:fwrite("Parsing file ~s, ", [InFile]),
	    case parse_head(Ifile, io:scan_erl_seq(Ifile, '', 1)) of
		{ok,REAs,Actions,Code} ->
		    io:fwrite("contained ~w rules.~n", [length(REAs)]),
		    file:close(Ifile),
		    {ok,REAs,Actions,Code};
		Error ->
		    file:close(Ifile),
		    Error
	    end;
	{error,R} ->
	    {error,{leex,{open,InFile}}}
    end.

%% parse_head(File, Toks)
%%  Parse the head of the file.

parse_head(Ifile, Toks) ->
    parse_defs(Ifile, Toks).

%% parse_defs(File, Toks)
%%  Parse the macro definition section of a file. Allow no definitions.

parse_defs(Ifile, {ok,[{var,Ld,'Definitions'},{dot,Ld}],Line}) ->
    parse_defs(Ifile, Line, []);
parse_defs(Ifile, Toks) ->
    parse_rules(Ifile, Toks, []).

parse_defs(Ifile, Line, Ms) ->
    parse_defs_1(Ifile, io:scan_erl_seq(Ifile, '', Line), Ms).

parse_defs_1(Ifile, {ok,[{var,Lm,M},{string,Le,Exp},{dot,Ld}],Line}, Ms) ->
    parse_defs(Ifile, Line, [{atom_to_list(M),Exp}|Ms]);
parse_defs_1(Ifile, {ok,[{atom,Lm,M},{string,Le,Exp},{dot,Ld}],Line}, Ms) ->
    parse_defs(Ifile, Line, [{atom_to_list(M),Exp}|Ms]);
parse_defs_1(Ifile, {ok,[{var,Lr,'Rules'},{dot,Ld}],Line}, Ms) ->
    parse_rules(Ifile, io:scan_erl_seq(Ifile, '', Line), Ms);
parse_defs_1(Ifile, {ok,Toks,Line}, Ms) ->
    {error,{Line,leex,bad_macro}};
parse_defs_1(Ifile, {error,E,Line}, Ms) ->
    {error,E};
parse_defs_1(Ifile, {eof,Line}, Ms) ->
    {error,{Line,leex,missing_rules}}.

%% parse_rules(File, Toks, Macros)
%%  Parse the RE rules section of the file. This must exist.

parse_rules(Ifile, Toks, Ms) ->
    parse_rules(Ifile, Toks, Ms, [], [], 0).

parse_rules(Ifile, {ok,[{string,Ls,S},{dot,Ld}],Line}, Ms, REAs, As, N) ->
    case parse_rule(S, Ls, empty_action, Ms, N) of
	{ok,REA,A} ->
	    parse_rules(Ifile, io:scan_erl_seq(Ifile, '', Line), Ms,
			[REA|REAs], [A|As], N+1);
	{error,E} -> {error,E}
    end;
parse_rules(Ifile, {ok,[{string,Ls,S}|Atoks],Line}, Ms, REAs, As, N) ->
    case parse_rule(S, Ls, Atoks, Ms, N) of
	{ok,REA,A} ->
	    parse_rules(Ifile, io:scan_erl_seq(Ifile, '', Line), Ms,
			[REA|REAs], [A|As], N+1);
	{error,E} -> {error,E}
    end;
parse_rules(Ifile, {ok,[{var,Le,'Erlang'},{atom,Lc,code},{dot,Ld}],Line},
	    Ms, REAs, As, N) ->
    %% Must be careful to put rules in correct order!
    parse_code(Ifile, Line, reverse(REAs), reverse(As));
parse_rules(Ifile, {ok,Toks,Line}, Ms, REAs, As, N) ->
    {error,{Line,leex,bad_rule}};
parse_rules(Ifile, {error,E,Line}, Ms, REAs, As, N) ->
    {error,E};
parse_rules(Ifile, {eof,Line}, Ms, REAs, As, N) ->
    %% Must be careful to put rules in correct order!
    {ok,reverse(REAs),reverse(As),[]}.

%% parse_rule(RegExpString, RegExpLine, ActionTokens, Macros, Counter)
%%  Parse one regexp after performing macro substition.

parse_rule(S, Line, empty_action, Ms, N) ->
    case parse_rule_regexp(S, Ms) of
	{regexp,R} ->
	    {ok,{R,N},{N,empty_action}};
	{error,E} ->
	    {error,{Line,leex,{regexp,S}}}
    end;
parse_rule(S, Line, Atoks, Ms, N) ->
    case parse_rule_regexp(S, Ms) of
	{regexp,R} ->
	    case erl_parse:parse_exprs(Atoks) of
		{ok,Aes} ->
		    YYtext = keymember('YYtext', 3, Atoks),
		    {ok,{R,N},{N,Aes,YYtext}};
		{error,E} ->
		    {error,{Line,leex,{after_regexp,S}}}
	    end;
	{error,E} ->
	    {error,{Line,leex,{regexp,S}}}
    end.

parse_rule_regexp(RE0, [{M,Exp}|Ms]) ->
    case old_string:re_gsub(RE0, [${|append(M, "}")], Exp) of
	{ok,RE} -> parse_rule_regexp(RE, Ms);
	{error,E} -> parse_rule_regexp(RE0, Ms)
    end;
parse_rule_regexp(RE, []) ->
    %%io:fwrite("RE = \"~s\"~n", [RE]),
    old_string:re_parse(RE).

%% parse_code(File, Line, REAs, Actions)
%%  Parse the code section of the file.

parse_code(Ifile, Line, REAs, As) ->
    {ok,REAs,As,io:get_chars(Ifile, '', 10240)}.

%% get_tokens(File)
%%  Scan one form from File. Return the list of tokens
%%  or 'error' or 'eof'.

get_tokens(File) ->
    case io:scan_erl_seq(File, '') of
	{ok,Toks,_} -> Toks;
	{error,E,_} -> error;
	{eof,_} -> eof
    end.

%% {NFA,FirstState} = build_combined_nfa(RegExpActionList)

build_combined_nfa(REAs) ->
    {NFA,Firsts,Free} = build_nfa_list(REAs, [], [], 0),
    {[{nfa,Free,epsilon_trans(Firsts),non_accepting}|NFA],Free}.

build_nfa_list([{RE,Action}|REAs], NFA0, Firsts, Free0) ->
    {NFA1,Free1,First} = build_nfa(RE, Free0, Action),
    build_nfa_list(REAs, append(NFA1, NFA0), [First|Firsts], Free1);
build_nfa_list([], NFA, Firsts, Free) ->
    {NFA,reverse(Firsts),Free}.

epsilon_trans([F|Fs]) ->
    [{epsilon,F}|epsilon_trans(Fs)];
epsilon_trans([]) ->
    [].

%% {NFA,NextFreeState,FirstState} = build_nfa(RegExp, Action)
%% {NFA,NextFreeState,FirstState} = build_nfa(RegExp, FreeState, Action)
%% {NewNFA,NextFreeState,FirstState} = build_nfa(RegExp, NFA, FreeState, NextState)

build_nfa(RE, Action) ->
    build_nfa(RE, 0, Action).

build_nfa({regexp,RE}, Free, Action) ->
    build_nfa(RE, [{nfa,Free,[],{accept,Action}}], Free+1, Free).

build_nfa([{span_0,RE1}|RE], NFA0, Free0, State0) ->
    {NFA1,Free1,State1} = build_nfa(RE, NFA0, Free0, State0),
    %% Reserve "place" for terminate and loop state.
    {NFA2,Free2,State2} = build_nfa(RE1, NFA1, Free1+1,Free1),
    {[{nfa,Free2,[{epsilon,State2},{epsilon,State1}],non_accepting},
      {nfa,Free1,[{epsilon,State2},{epsilon,State1}],non_accepting}|NFA2],
     Free2+1,Free2};
build_nfa([{span_1,RE1}|RE], NFA0, Free0, State0) ->
    {NFA1,Free1,State1} = build_nfa(RE, NFA0, Free0, State0),
    %% Reserve "place" for terminate and loop state.
    {NFA2,Free2,State2} = build_nfa(RE1, NFA1, Free1+1,Free1),
    {[{nfa,Free2,[{epsilon,State2}],non_accepting},
      {nfa,Free1,[{epsilon,State2},{epsilon,State1}],non_accepting}|NFA2],
     Free2+1,Free2};
build_nfa([{span_0_1,RE1}|RE], NFA0, Free0, State0) ->
    {NFA1,Free1,State1} = build_nfa(RE, NFA0, Free0, State0),
    %% Reserve "place" for terminate state.
    {NFA2,Free2,State2} = build_nfa(RE1, NFA1, Free1+1,Free1),
    {[{nfa,Free2,[{epsilon,State2},{epsilon,State1}],non_accepting},
      {nfa,Free1,[{epsilon,State1}],non_accepting}|NFA2],
     Free2+1,Free2};
build_nfa([{group,RE1}|RE], NFA0, Free0, State0) ->
    {NFA1,Free1,State1} = build_nfa(RE, NFA0, Free0, State0),
    build_nfa(RE1, NFA1, Free1, State1);
build_nfa([{alt,Lhs,Rhs}|RE], NFA0, Free0, State0) ->
    {NFA1,Free1,State1} = build_nfa(RE, NFA0, Free0, State0),
    %% E1 and E2 are states with e-transitions to next state.
    E1 = {nfa,Free1,[{epsilon,State1}],non_accepting},
    {NFA2,Free2,State2} = build_nfa(Lhs, [E1|NFA1], Free1+1, Free1),
    E2 = {nfa,Free2,[{epsilon,State1}],non_accepting},
    {NFA3,Free3,State3} = build_nfa(Rhs, [E2|NFA2], Free2+1, Free2),
    {[{nfa,Free3,[{epsilon,State2},{epsilon,State3}],non_accepting}|NFA3],
     Free3+1,Free3};
build_nfa([{compl_set,Set0}|RE], NFA0, Free0, State0) ->
    {NFA1,Free1,State1} = build_nfa(RE, NFA0, Free0, State0),
    Set = nfa_compl_set(Set0),
    {NFA2,Free2,Trans} = nfa_set_states(Set, NFA1, Free1, State1),
    {[{nfa,Free2,Trans,non_accepting}|NFA2],Free2+1,Free2};
build_nfa([{char_set,Set0}|RE], NFA0, Free0, State0) ->
    {NFA1,Free1,State1} = build_nfa(RE, NFA0, Free0, State0),
    Set = nfa_char_set(Set0),
    {NFA2,Free2,Trans} = nfa_set_states(Set, NFA1, Free1, State1),
    {[{nfa,Free2,Trans,non_accepting}|NFA2],Free2+1,Free2};
build_nfa([char|RE], NFA, Free, State) ->
    build_nfa([{compl_set,[$\r,$\n]}|RE], NFA, Free, State);
build_nfa([C|RE], NFA0, Free0, State0) when C >= 0, C < 256 ->
    {NFA1,Free1,State1} = build_nfa(RE, NFA0, Free0, State0),
    {[{nfa,Free1,[{C,State1}],non_accepting}|NFA1],Free1+1,Free1};
build_nfa([epsilon|RE], NFA, Free, State) ->    %The empty string/expression
    build_nfa(RE, NFA, Free, State);
build_nfa([], NFA, Free, State) ->
    {NFA,Free,State}.

nfa_set_states([C|Set], NFA0, Free0, State) ->
    {NFA1,Free1,Trans} = nfa_set_states(Set, NFA0, Free0, State),
    {[{nfa,Free1+1,[{C,Free1}],non_accepting},
      {nfa,Free1,[{epsilon,State}],non_accepting}|NFA1],
     Free1+2,[{epsilon,Free1+1}|Trans]};
nfa_set_states([], NFA, Free, State) ->
    {NFA,Free,[]}.

nfa_char_set(Chars) ->
    nfa_pack_set(nfa_char_set(Chars, [])).

nfa_compl_set(Chars) ->
    nfa_pack_set(subtract(nfa_char_range(0,255),nfa_char_set(Chars, []))).

nfa_char_set([{Cf,Cl}|Chars], Set) ->
    nfa_char_set(Chars, union(nfa_char_range(Cf, Cl), Set));
nfa_char_set([C|Chars], Set) ->
    nfa_char_set(Chars, add_element(C, Set));
nfa_char_set([], Set) ->
    Set.

nfa_char_range(Last, Last) ->
    [Last];
nfa_char_range(C, Last) ->
    [C|nfa_char_range(C+1, Last)].

nfa_pack_set([C|Cs]) ->
    nfa_pack_set(Cs, C, C).

nfa_pack_set([C|Cs], Cf, Cl) when C == Cl + 1 ->
    nfa_pack_set(Cs, Cf, C);
nfa_pack_set([C|Cs], Cf, Cf) ->
    [Cf|nfa_pack_set(Cs, C, C)];
nfa_pack_set([C|Cs], Cf, Cl) ->
    [{range,Cf,Cl}|nfa_pack_set(Cs, C, C)];
nfa_pack_set([], C, C) ->
    [C];
nfa_pack_set([], Cf, Cl) ->
    [{range,Cf,Cl}].

%% nfa_trans(State, NFA)
%% nfa_action(State, NFA)
%%  Access the transaction and action field of an NFA state.

nfa_trans(State, [{nfa,State,Trans,Action}|NFA]) ->
    Trans;
nfa_trans(State, [_|NFA]) ->
    nfa_trans(State, NFA).

nfa_action(State, [{nfa,State,Trans,Action}|NFA]) ->
    Action;
nfa_action(State, [_|NFA]) ->
    nfa_action(State, NFA).

%% {DFA,FirstState} = build_dfa(NFA, FirstState)

build_dfa(NFA, First) ->
    build_dfa(NFA, 0, First).

build_dfa(NFA, Free, First) ->
    {DFA,Free1,DFASts,First1} = build_dfa([First], [], Free, [], NFA),
    {DFA,First1}.

build_dfa(Sts, DFA0, Free0, DFASts, NFA) ->
    EClosure = e_closure(Sts, NFA),
    Action = action(EClosure, NFA),
    Ics = input_chars(EClosure, NFA),
    {DFA1,Free1,DFASts1,Trans} = build_dfa_trans(Ics, EClosure, DFA0, Free0+1,
						 [{Sts,Free0}|DFASts], NFA),
    {[{dfa,Free0,Sts,EClosure,Trans,Action}|DFA1],Free1,DFASts1,Free0}.

build_dfa_trans([Ic|Ics], Closure, DFA0, Free0, DFASts, NFA) ->
    case move(Closure, Ic, NFA) of
	[] -> build_dfa_trans(Ics, Closure, DFA0, Free0, DFASts, NFA);
	Sts ->
	    {DFA1,Free1,DFASts1,St} = build_dfa_cont(Sts, DFA0, Free0, DFASts, NFA),
	    {DFA2,Free2,DFASts2,Trans} = build_dfa_trans(Ics, Closure, DFA1, Free1, DFASts1, NFA),
	    {DFA2,Free2,DFASts2,[{Ic,St}|Trans]}
    end;
build_dfa_trans([], Closure, DFA, Free, DFASts, NFA) ->
    {DFA,Free,DFASts,[]}.

build_dfa_cont(Sts, DFA0, Free0, DFASts, NFA) ->
    case dfa_states_state(Sts, DFASts) of
	{state,St} ->
	    {DFA0,Free0,DFASts,St};
	new ->
	    build_dfa(Sts, DFA0, Free0, DFASts, NFA)
    end.

dfa_states_state(Sts, [{Sts,N}|DFASts]) ->
    {state,N};
dfa_states_state(Sts, [_|DFASts]) ->
    dfa_states_state(Sts, DFASts);
dfa_states_state(Sts, []) ->
    new.

%% input_chars(States, NFA)
%%  Return the set of all the input characters from States.

input_chars(States, NFA) ->
    input_chars(States, NFA, []).

input_chars([S|Ss], NFA, Ics) ->
    input_chars(Ss, NFA, union(input_chars1(nfa_trans(S, NFA)), Ics));
input_chars([], NFA, Ics) ->
    Ics.

input_chars1([{epsilon,S}|Ts]) -> [];           %Only e-transitions
input_chars1([{{range,Cf,Cl},S}]) -> nfa_char_range(Cf, Cl);
input_chars1([{C,N}]) -> [C];
input_chars1([]) -> [].

%% E-Closure = e_closure(States, NFA)
%%  Return the set of states in the e-closure of States in NFA.

e_closure(T, NFA) ->
    e_closure(T, NFA, []).

e_closure([S|T], NFA, Closure) ->
    case is_element(S, Closure) of
	true -> e_closure(T, NFA, Closure);
	false ->
	    e_closure(append(epsilon_states(S, NFA), T),
		      NFA, add_element(S, Closure))
    end;
e_closure([], NFA, Closure) ->
    Closure.

epsilon_states(State, NFA) ->
    epsilon_states(nfa_trans(State, NFA)).

epsilon_states([{epsilon,N}|Ts]) ->
    [N|epsilon_states(Ts)];
epsilon_states([E|Ts]) ->
    epsilon_states(Ts);
epsilon_states([]) ->
    [].

%% move(States, InputChar, NFA)
%%  Return all the set of states that can be reached on InputChar from
%%  any state in States.

move(States, Char, NFA) ->
    move(States, Char, NFA, []).

move([S|Ss], Ic, NFA, MS0) ->
    MS = move1(nfa_trans(S, NFA), Ic, MS0),
    move(Ss, Ic, NFA, MS);
move([], Ic, NFA, MS) ->
    MS.

move1([{{range,Cf,Cl},S}|Ts], Ic, MS) when Ic >= Cf, Ic =< Cl ->
    move1(Ts, Ic, add_element(S, MS));
move1([{Ic,S}|Ts], Ic, MS) ->
    move1(Ts, Ic, add_element(S, MS));
move1([T|Ts], Ic, MS) ->
    move1(Ts, Ic, MS);
move1([], Ic, MS) ->
    MS.

%% action(States, NFA)
%%  Return the *first* significant action from the set of states States.
%%  Assume States is an ordered set.

action([S|Ss], NFA) ->
    case nfa_action(S, NFA) of
	non_accepting -> action(Ss, NFA);
	Action -> Action
    end;
action([], NFA) -> non_accepting.               %The default is no action

%% minimise_dfa(DFA, FirstState)
%%  Minimise the DFA by removing equivalent states. Our algorithm is
%%  different from the one in the Dragon book, but equivalent, I think.
%%  We scan through all the states and remove duplicate states with
%%  exactly the same transitions and actions, then all the transitions
%%  are updated. This procedure is then repeated until no states are
%%  removed. At the same time we compact the state numbers.

%%minimise_dfa(DFA, DF) ->
%%    {DFA,DF}.
minimise_dfa(DFA0, DF0) ->
    case min_dfa(DFA0, 0, [], []) of
	{DFA,Reds} when length(DFA) == length(DFA0) ->  %No reduction
	    {DFA0,DF0};
	{DFA1,Reds} ->
	    DFA = min_update(DFA1, Reds),
	    minimise_dfa(DFA, min_use(DF0, Reds))
    end.

min_dfa([{dfa,S,Sts,Cl,Trs,Action}|DFA0], NewS, Reds0, MDFA) ->
    {DFA,Reds} = min_unique(DFA0, NewS, Trs, Action, [], [{S,NewS}|Reds0]),
    min_dfa(DFA, NewS+1, Reds, [{dfa,NewS,Sts,Cl,Trs,Action}|MDFA]);
min_dfa([], NewS, Reds, MDFA) ->
    {reverse(MDFA),Reds}.

min_unique([{dfa,S,Sts,Cl,Trs,Action}|DFAs], UseS, Trs, Action, RDFA, Reds) ->
    min_unique(DFAs, UseS, Trs, Action, RDFA, [{S,UseS}|Reds]);
min_unique([State|DFAs], UseS, Trs, Action, RDFA, Reds) ->
    min_unique(DFAs, UseS, Trs, Action, [State|RDFA], Reds);
min_unique([], UseS, Trs, Action, RDFA, Reds) ->
    {reverse(RDFA),Reds}.

min_update([{dfa,S,Sts,Cl,Trs,Action}|DFA], Reds) ->
    [{dfa,S,Sts,Cl,min_update_trans(Trs, Reds),Action}|min_update(DFA, Reds)];
min_update([], Reds) -> [].

min_update_trans([{C,S}|Trs], Reds) ->
    [{C,min_use(S, Reds)}|min_update_trans(Trs, Reds)];
min_update_trans([], Reds) -> [].

min_use(Old, [{Old,New}|Reds]) -> New;
min_use(Old, [R|Reds]) -> min_use(Old, Reds);
min_use(Old, []) -> Old.

%% out_file(OutFile, OutName, DFA, FirstState, Actions, IncludedCode)
%%  Generate the scanner file. Put the included code directly after the
%%  prologue to allow the user to export and import functions.
%%  The function state/4 tries to scan one token by calling itself
%%  recursively until a token has been scanned, no more can be scanned
%%  on the token started or input has ended. The function calling
%%  state/4 must interpret the return values correctly in its context.
%%  State/4 returns:
%%
%%  {accept,RestInChars,TokenChars,Action}
%%	A complete token has definitely been scanned.
%%
%%  {AcceptState,CurrentState,RestInChars,TokenChars}
%%	A token could not definitely be scanned, RestInChars will show if
%%	there was an illegal character or end of input was reached.
%%	AcceptState can be:
%%
%%	{RestInChars,TokenChars,Action}
%%		- a token was scanned this far but we could go further.
%%	reject	- no token could be scanned at all.

out_file(OutFile, Out, DFA, DF, Actions, Code) ->
    io:fwrite("Writing file ~s, ", [OutFile]),
    case file:open(OutFile, write) of
	{ok,Ofile} ->
	    out_prologue(Ofile, Out),
	    out_code(Ofile, Code),
	    out_include(Ofile),
	    out_dfa(Ofile, DFA, DF),
	    out_actions(Ofile, Actions),
	    file:close(Ofile),
	    io:fwrite("ok~n");
	{error,R} ->
	    io:fwrite("open error~n"),
	    error
    end.

%% out_prologue(File, Out)

out_prologue(File, Out) ->
    io:fwrite(File, "-module(~s).~n~n", [filename:basename(Out)]),
    io:fwrite(File, "-export([string/1,string/2,token/2,token/3,tokens/2,tokens/3]).~n~n",
	      []).

%% out_code(File, Code)
%%  Add the included code.

out_code(File, Code) ->
    io:fwrite(File, "%% Included code.~n", []),
    io:put_chars(File, [Code,$\n]).

%% out_include(File)
%%  Add the standard include file.

out_include(File) ->
%tobbe    case file:path_open([".", [code:lib_dir(),"/parsetools/include"]],
%tobbe			 "leex.hrl", read) of
    case file:open("leex.hrl", read) of
	{ok,Ifile} -> 
	    io:put_chars(File, io:get_chars(Ifile, '', 10240)),
	    io:nl(File),
	    file:close(Ifile);
	{error,E} ->
	    io:fwrite(File, "%% No standard functions.~n~n", [])
    end.

%% out_dfa(File, DFA, DF)

out_dfa(File, DFA, DF) ->
    io:fwrite(File, "%% Generated DFA tables and actions.~n", []),
    io:fwrite(File, "yystate() -> ~w.~n~n", [DF]),
    io:fwrite(File, "%% yystate(State, InChars, LineNo, TokenChars, AcceptState)~n~n", []),
    out_states(File, DFA).

out_states(File, [{dfa,State,_,_,Trans0,non_accepting}|Ss]) ->
    Trans = keysort(1, out_pack_trans(Trans0)),
    out_na_state(File, State, Trans),
    out_states(File, Ss);
out_states(File, [{dfa,State,_,_,[],{accept,Action}}|Ss]) ->
    io:fwrite(File, "yystate(~w, Ics, Line, Tcs, Accept) ->\n", [State]),
    io:fwrite(File, "    {accept,Ics,Line,Tcs,~w};\n", [Action]),
    out_states(File, Ss);
out_states(File, [{dfa,State,_,_,Trans0,{accept,Action}}|Ss]) ->
    Trans = keysort(1, out_pack_trans(Trans0)),	%Bunch together
    out_a_state(File, State, Trans, Action),
    out_states(File, Ss);
out_states(File, []) ->
    io:fwrite(File, "yystate(_, _, _, _, _) -> error.~n~n", []).

out_na_state(File, State, [{{range,Cf,Cl},Next}|Ts]) ->
    out_head(File, State, io_lib:write_char(Cf), io_lib:write_char(Cl)),
    out_na_body(File, Next, "Line", "C"),
    out_na_state(File, State, Ts);
out_na_state(File, State, [{$\n,Next}|Ts]) ->
    out_head(File, State, "$\\n"),
    out_na_body(File, Next, "Line+1", "$\\n"),
    out_na_state(File, State, Ts);
out_na_state(File, State, [{C,Next}|Ts]) ->
    out_head(File, State, io_lib:write_char(C)),
    out_na_body(File, Next, "Line", io_lib:write_char(C)),
    out_na_state(File, State, Ts);
out_na_state(File, State, []) ->
    io:fwrite(File, "yystate(~w, Ics, Line, Tcs, Accept) ->\n", [State]),
    io:fwrite(File, "    {Accept,~w,Ics,Line,Tcs};\n", [State]).

out_a_state(File, State, [{{range,Cf,Cl},Next}|Ts], Action) ->
    out_head(File, State, io_lib:write_char(Cf), io_lib:write_char(Cl)),
    out_a_body(File, Next, "Line", "C", Action),
    out_a_state(File, State, Ts, Action);
out_a_state(File, State, [{$\n,Next}|Ts], Action) ->
    out_head(File, State, "$\\n"),
    out_a_body(File, Next, "Line+1", "$\\n", Action),
    out_a_state(File, State, Ts, Action);
out_a_state(File, State, [{C,Next}|Ts], Action) ->
    out_head(File, State, io_lib:write_char(C)),
    out_a_body(File, Next, "Line", io_lib:write_char(C), Action),
    out_a_state(File, State, Ts, Action);
out_a_state(File, State, [], Action) ->
    io:fwrite(File, "yystate(~w, Ics, Line, Tcs, Accept) ->\n", [State]),
    io:fwrite(File, "    {{Ics,Tcs,~w},~w,Ics,Line,Tcs};\n", [Action,State]).

out_head(File, State, C) ->
    io:fwrite(File, "yystate(~w, [~s|Ics], Line, Tcs, Accept) ->\n", [State,C]).

out_head(File, State, Min, Max) ->
    io:fwrite(File, "yystate(~w, [C|Ics], Line, Tcs, Accept) when C >= ~s, C =< ~s ->\n",
	      [State,Min,Max]).

out_na_body(File, Next, Line, C) ->
    io:fwrite(File, "    yystate(~w, Ics, ~s, [~s|Tcs], Accept);\n", [Next,Line,C]).

out_a_body(File, Next, Line, C, Action) ->
    io:fwrite(File, "    yystate(~w, Ics, ~s, [~s|Tcs], {[~s|Ics],Tcs,~w});\n",
	      [Next,Line,C,C,Action]).

out_pack_trans([{$\n,S}|Ts]) ->
    [{$\n,S}|out_pack_trans(Ts)];
out_pack_trans([{C,S}|Ts]) ->
    out_pack_trans(Ts, C, C, S);
out_pack_trans([]) -> [].

out_pack_trans([{$\n,S}|Ts], Cf, Cl, S1) ->
    out_pack_trans_1(Cf, Cl, S1, [{$\n,S}|out_pack_trans(Ts)]);
out_pack_trans([{C,S}|Ts], Cf, Cl, S) when C == Cl + 1 ->
    out_pack_trans(Ts, Cf, C, S);
out_pack_trans([{C,S}|Ts], Cf, Cl, S1) ->
    out_pack_trans_1(Cf, Cl, S1, out_pack_trans(Ts, C, C, S));
out_pack_trans([], Cf, Cl, S) ->
    out_pack_trans_1(Cf, Cl, S, []).

out_pack_trans_1(C, C, S, Rest) ->
    [{C,S}|Rest];
out_pack_trans_1(Cf, Cl, S, Rest) when Cl == Cf + 1 ->
    [{Cf,S},{Cl,S}|Rest];
%%out_pack_trans_1(Cf, Cl, S, Rest) when Cl == Cf + 2 ->
%%    [{Cf,S},{Cf+1,S},{Cl,S}|Rest];
out_pack_trans_1(Cf, Cl, S, Rest) ->
    [{{range,Cf,Cl},S}|Rest].

out_actions(File, [{A,empty_action}|As]) ->
    io:fwrite(File, "yyaction(~w, YYrevtext, YYline, YYts) -> YYts;~n", [A]),
    out_actions(File, As);
out_actions(File, [{A,Code,YYtext}|As]) ->
    io:fwrite(File, "yyaction(~w, YYrevtext, YYline, YYts) ->~n", [A]),
    if
	YYtext == true ->
	    io:fwrite(File, "    YYtext = yyrev(YYrevtext),~n", []);
	YYtext == false -> ok
    end,
    io:fwrite(File, "    [begin ~s end|YYts];~n", [erl_pp:seq(Code)]),
    out_actions(File, As);
out_actions(File, []) ->
    io:fwrite(File, "yyaction(_, _, _, _) -> error.~n", []).

%%% Interface to erl_compile.

compile(Input, Output, _Options) ->
    gen(Input, Output).
